/*
 * dstumbler v1.0 [mon.c]
 * by h1kari - (c) Dachb0den Labs 2001
 */

/*
 * Copyright (c) 2001 Dachb0den Labs.
 *      David Hulton <h1kari@dachb0den.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by David Hulton.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY David Hulton AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL David Hulton OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <ctype.h>

#include <pcap.h>
#include <net/bpf.h>
#define	IEEE80211_CHAN_MAX	255
#include <net80211/ieee80211.h>

#include "dstumbler.h"
#include "config.h"
#include "screen.h"

pcap_t *p;
u_char *buf;
struct pcap_pkthdr h;
u_char chan_avail[roundup(IEEE80211_CHAN_MAX,NBBY)];
int monswspeed = MONSWSPEED_DEF;

void
mon_adjustspeed(int increment)
{
   if (increment)
   {
     if (monswspeed < MONSWSPEED_MAX)
       monswspeed += 100*1000;
   }
   else
   {
     if (monswspeed > MONSWSPEED_MIN)
       monswspeed -= 100*1000;
   }
   notify("speed %u ms", monswspeed/1000);
}

/*
 * just grab the next packet so we can process it in both mon_parse_ap and
 * mon_parse_node
 */
int
mon_next(const char *iface)
{
  static int first = 1;
  static struct timeval now, then;

  if(monmode && !chanlock && (CMPTVAL(now, then) > -1 || first))
  {
again:
    do {
      ch %= SWITCH_CHANS;
      ch++;
    } while (isclr(chan_avail, ch));
    if (setdebugchan(iface, ch) != 0)
    {
	/* invalid channel, don't try it again */
	clrbit(chan_avail, ch);
	goto again;
    }
    notify("channel %u", ch);

    gettimeofday(&now, NULL);
    ADDTVAL(now, then, monswspeed);

    first = 0;
  }

  buf = (u_char *)pcap_next(p, &h);
  return (buf == NULL ? -1 : 0);
}

#define	IEEE80211_ADDR_EQ(a1,a2)	(memcmp(a1,a2,IEEE80211_ADDR_LEN) == 0)
#define	IEEE80211_ADDR_COPY(dst,src)	memcpy(dst,src,IEEE80211_ADDR_LEN)

/*
 * parse buf for a beacon packet or proberesponse so we can get ap info
 */
int
mon_parse_ap(const char *iface, struct aps_s *aps)
{
  struct ieee80211_frame *wh;
  u_int8_t *frm, *efrm;
  u_int8_t *ssid, *rates, *xrates;
  u_int8_t *tstamp, *intval, *capinfo;
  u_int8_t chan, subtype;
  u_int16_t v, algo, seq;
  int i, len;

  if(buf == NULL)
    return -1;

  wh = (struct ieee80211_frame *)buf;
  if ((wh->i_fc[0] & IEEE80211_FC0_VERSION_MASK) != IEEE80211_FC0_VERSION_0)
    return -1;
  if ((wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK) != IEEE80211_FC0_TYPE_MGT)
    return -1;
  subtype = wh->i_fc[0] & IEEE80211_FC0_SUBTYPE_MASK;
  frm = (u_char *)&wh[1];
  efrm = buf + h.caplen;
  switch (subtype)
  {
    case IEEE80211_FC0_SUBTYPE_BEACON:
    case IEEE80211_FC0_SUBTYPE_PROBE_RESP:
	  if (efrm - frm < 12)
	  {
	    notify("short beacon/probe-resp (%u)", efrm-frm);
	    return -1;
	  }
	  tstamp = frm; frm += 8;
	  intval = frm; frm += 2;
	  capinfo = frm; frm += 2;
	  ssid = rates = xrates = NULL;
	  chan = 0;

	  while(frm < efrm) 
	  {
	    switch(*frm)
	    {
	      case IEEE80211_ELEMID_SSID:
		ssid = frm;
		break;
	      case IEEE80211_ELEMID_RATES:
		rates = frm;
		break;
	      case IEEE80211_ELEMID_XRATES:
		xrates = frm;
		break;
	      case IEEE80211_ELEMID_DSPARMS:
		chan = frm[2];
		break;
	    }
	    frm += frm[1] + 2;
	  }

	  if(chan == 0 || !ssid || !rates)
	  {
	    notify("missing element id");
	    return -1;
	  }

	  IEEE80211_ADDR_COPY(aps->bssid, wh->i_addr2);

	  len = MIN(ssid[1], SSIDLEN), ssid += 2;
	  for (i = 0; i < len; i++)
	  {
	    if (isprint(ssid[i]))
	      aps->ssid[i] = ssid[i];
	    else
	      aps->ssid[i] = '?';
	  }
	  aps->ssid[i] = '\0';

	  len = rates[1], rates += 2;
	  for (; len > 0; len--, rates++)
	    aps->srate = MAX(aps->srate, (rates[0] & IEEE80211_RATE_VAL));
	  if (xrates)
	  {
	    len = xrates[1], xrates += 2;
	    for (; len > 0; len--, xrates++)
	      aps->srate = MAX(aps->srate, (xrates[0] & IEEE80211_RATE_VAL));
	  }

	  aps->chan = chan;

	  memcpy(&v, capinfo, sizeof(u_int16_t));
          aps->weped = (le16toh(v) & IEEE80211_CAPINFO_PRIVACY) ?  1 : 2;
          aps->adhoc = (le16toh(v) & IEEE80211_CAPINFO_IBSS) ?  1 : 2;
	  memcpy(&v, intval, sizeof(u_int16_t));
          aps->interval = le16toh(v);
#if 0
	  /* XXX no rssi info */
	  aps->quality[2] = MAX(wi_h->wi_silence, 27);
	  aps->quality[1] = MAX(wi_h->wi_signal, 27);
	  aps->quality[0] = MAX(aps->quality[1] - aps->quality[2], 0);
#else
	  aps->quality[2] = wh->i_dur[1];
	  aps->quality[1] = wh->i_dur[0];
	  aps->quality[0] = aps->quality[1];
#endif
	  return 0;
    case IEEE80211_FC0_SUBTYPE_AUTH:
	  if (efrm - frm < 6)
	  {
	    notify("short auth frame (%u)", efrm-frm);
	    return -1;
	  }
	  memcpy(&algo, frm, sizeof(u_int16_t)), frm += sizeof(u_int16_t);
	  memcpy(&seq, frm, sizeof(u_int16_t)), frm += sizeof(u_int16_t);
          if(!(le16toh(seq) == 2 || le16toh(seq) == 4))
            return -1;

          /* set weped to on so we don't disturb the other settings */
          aps->weped = 1;
          aps->keyed = le16toh(algo) ? 1 : 2; 

          return 0;
  }
  return -1;
}

/*
 * parse packet for possible network node information
 */
#define NODE_BROADCAST "\xff\xff\xff\xff\xff\xff"
int
mon_parse_node(struct node_s *node, char *bssid)
{
  struct ieee80211_frame *wh;
  u_int8_t *frm, *efrm;
  u_int8_t *rates;
  u_int8_t *intval, *capinfo;
  u_int8_t subtype;
  u_int16_t v, algo;
  int len;

  if(buf == NULL)
    return -1;

  wh = (struct ieee80211_frame *)buf;
  if ((wh->i_fc[0] & IEEE80211_FC0_VERSION_MASK) != IEEE80211_FC0_VERSION_0)
    return -1;
  subtype = wh->i_fc[0] & IEEE80211_FC0_SUBTYPE_MASK;
  /* Ignore beacons, they're broadcast */
  if ((wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK) == IEEE80211_FC0_TYPE_MGT &&
     subtype == IEEE80211_FC0_SUBTYPE_BEACON)
    return -1;
  switch (wh->i_fc[1] & IEEE80211_FC1_DIR_MASK) {
  case IEEE80211_FC1_DIR_NODS:
    IEEE80211_ADDR_COPY(bssid, wh->i_addr3);
    IEEE80211_ADDR_COPY(node->mac, wh->i_addr2);

    if(IEEE80211_ADDR_EQ(node->mac, bssid))
      IEEE80211_ADDR_COPY(node->mac, wh->i_addr1);
    break;
  case IEEE80211_FC1_DIR_TODS:
    IEEE80211_ADDR_COPY(bssid, wh->i_addr1);
    IEEE80211_ADDR_COPY(node->mac, wh->i_addr2);
    break;
  case IEEE80211_FC1_DIR_FROMDS:
    IEEE80211_ADDR_COPY(bssid, wh->i_addr2);
    IEEE80211_ADDR_COPY(node->mac, wh->i_addr1);
    break;
  case IEEE80211_FC1_DIR_DSTODS:
    return -1;
  }
  /* if we got a broadcast address, this packet is junk */
  if(subtype != IEEE80211_FC0_SUBTYPE_PROBE_REQ &&
   (IEEE80211_ADDR_EQ(bssid, NODE_BROADCAST) ||
    IEEE80211_ADDR_EQ(node->mac, NODE_BROADCAST) ||
    (node->mac[0] != 0x0 && node->mac[1] != 0x8)))
    return -1;
#if 0
  node->quality[2] = MAX(wi_h->wi_silence, 27);
  node->quality[1] = MAX(wi_h->wi_signal, 27);
  node->quality[0] = MAX(node->quality[1] - node->quality[2], 0);
#else
  node->quality[2] = wh->i_dur[1];
  node->quality[1] = wh->i_dur[0];
  node->quality[0] = node->quality[1];
#endif
  switch(wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK)
  {
    case IEEE80211_FC0_TYPE_DATA:
      /* if this is a data packet, set the amount of data sent */
      node->datalen = h.len;
      node->keyid = *(u_char *)(buf + WEP_KEYID_OFF);
      node->keyid = (node->keyid & 0xc0) >> 6;
      break;
    case IEEE80211_FC0_TYPE_MGT:
      frm = (u_char *)&wh[1];
      efrm = buf + h.caplen;
      switch(subtype)
      {
        case IEEE80211_FC0_SUBTYPE_PROBE_REQ:
        case IEEE80211_FC0_SUBTYPE_ASSOC_REQ:
        case IEEE80211_FC0_SUBTYPE_REASSOC_REQ:
          if(subtype == IEEE80211_FC0_SUBTYPE_ASSOC_REQ)
          {
	    if (efrm - frm != 4)
		return -1;
	    capinfo = frm; frm += 2;
	    intval = frm; frm += 2;
          }
          else if(subtype == IEEE80211_FC0_SUBTYPE_REASSOC_REQ)
          {
	    if (efrm - frm != 10)
		return -1;
	    capinfo = frm; frm += 2;
	    intval = frm; frm += 2;
	    frm += 6;		/* ignore current AP info */
          }

	  memcpy(&v, capinfo, sizeof(u_int16_t));
          node->weped = (le16toh(v) & IEEE80211_CAPINFO_PRIVACY) ?  1 : 2;
          if(le16toh(v) & IEEE80211_CAPINFO_IBSS)
		node->adhoc = 1;
          if(le16toh(v) & IEEE80211_CAPINFO_ESS)
		node->adhoc = 2;
	  memcpy(&v, intval, sizeof(u_int16_t));
          node->interval = le16toh(v);

	  while(frm < efrm) 
	  {
	    switch(*frm)
	    {
	      case IEEE80211_ELEMID_SSID:
                if(subtype == IEEE80211_FC0_SUBTYPE_PROBE_REQ)
                  node_setany(node->mac, frm[1] ? 0 : 1);
		break;
	      case IEEE80211_ELEMID_RATES:
		  rates = frm;
		  len = rates[1], rates += 2;
		  for (; len > 0; len--, rates++)
			node->srate = MAX(node->srate,
				(rates[0] & IEEE80211_RATE_VAL));
                if(subtype == IEEE80211_FC0_SUBTYPE_PROBE_REQ)
                  node_setsrate(node->mac, node->srate);
		break;
	    }
	    frm += frm[1] + 2;
	  }

          if(subtype == IEEE80211_FC0_SUBTYPE_PROBE_REQ)
            return -1;

          break;
        case IEEE80211_FC0_SUBTYPE_AUTH:
	  memcpy(&algo, frm, sizeof(u_int16_t));
          if(le16toh(algo))
            node->keyed = 1;
          else
            node->keyed = 2;
          break;
      }
      break;
  }

  return 0;
}

static int
mon_checkfordlt(pcap_t *p, int dlt)
{
  int i, n, *dl;
  int found = 0;

  n = pcap_list_datalinks(p, &dl);
  for(i = 0; i < n; i++)
	if (dl[i] == dlt) {
		found = 1;
		break;
	}
  free(dl);
  return found;
}

void
mon_start(char *iface)
{
  char errbuf[PCAP_ERRBUF_SIZE];

  p = pcap_open_live(iface, PCAP_SNAPLEN, PCAP_PROMISC, PCAP_TOUT, errbuf);
  if(p == NULL)
  {
    alert("error: unable to open pcap device: %s", strerror(errno));
    exitclean(2);
  }
  if(!mon_checkfordlt(p, DLT_IEEE802_11))
  {
    alert("error: interface does not support an IEEE802_11 DLT");
    exitclean(2);
  }
  pcap_set_datalink(p, DLT_IEEE802_11);
  memset(chan_avail, 0xff, sizeof(chan_avail));

  buf = NULL;
}

void
mon_stop(void)
{
  if(p != NULL)
    pcap_close(p);
}
